#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import re
from typing import List

import requests

from airbyte_cdk.sources.declarative.extractors.record_extractor import RecordExtractor
from airbyte_cdk.sources.declarative.types import Record
from airbyte_cdk.sources.streams.http.error_handlers import BackoffStrategy


class AdAccountRecordExtractor(RecordExtractor):
    """
    Custom extractor for handling different response formats from the Ad Accounts endpoint.

    This extractor is necessary to handle cases where an `account_id` is present in the request.
    - When querying all ad accounts, the response contains an "items" key with a list of accounts.
    - When querying a specific ad account, the response returns a single dictionary representing that account.
    """

    def extract_records(self, response: requests.Response) -> List[Record]:
        data = response.json()

        if not data:
            return []

        # Extract records from "items" if present
        if isinstance(data, dict) and "items" in data:
            return data["items"]

        # If the response is a single object, wrap it in a list
        if isinstance(data, dict):
            return [data]
        return []


class PinterestAnalyticsBackoffStrategy(BackoffStrategy):
    _re = re.compile(r"Retry after\s+(\d+)\s+seconds", re.IGNORECASE)

    def backoff_time(self, response_or_exception, attempt_count: int) -> float:
        try:
            if isinstance(response_or_exception, requests.Response):
                data = response_or_exception.json()
                msg = str(data.get("message", ""))
                m = self._re.search(msg)
                if m:
                    return float(m.group(1))
        except Exception:
            pass
        return min(2**attempt_count, 120.0)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from datetime import datetime
from functools import cache


@cache
def get_analytics_columns() -> str:
    # https://developers.pinterest.com/docs/api/v5/#operation/ad_account/analytics
    analytics_columns = [
        "ADVERTISER_ID",  #
        "AD_ACCOUNT_ID",
        "AD_GROUP_ENTITY_STATUS",
        "AD_GROUP_ID",
        "AD_ID",
        "CAMPAIGN_DAILY_SPEND_CAP",
        "CAMPAIGN_ENTITY_STATUS",
        "CAMPAIGN_ID",
        "CAMPAIGN_LIFETIME_SPEND_CAP",
        "CAMPAIGN_NAME",
        "CHECKOUT_ROAS",
        "CLICKTHROUGH_1",  #
        "CLICKTHROUGH_1_GROSS",  #
        "CLICKTHROUGH_2",  #
        "CPC_IN_MICRO_DOLLAR",
        "CPM_IN_DOLLAR",
        "CPM_IN_MICRO_DOLLAR",
        "CTR",
        "CTR_2",
        "ECPCV_IN_DOLLAR",
        "ECPCV_P95_IN_DOLLAR",
        "ECPC_IN_DOLLAR",
        "ECPC_IN_MICRO_DOLLAR",
        "ECPE_IN_DOLLAR",
        "ECPM_IN_MICRO_DOLLAR",
        "ECPV_IN_DOLLAR",
        "ECTR",
        "EENGAGEMENT_RATE",
        "ENGAGEMENT_1",  #
        "ENGAGEMENT_2",  #
        "ENGAGEMENT_RATE",
        "IDEA_PIN_PRODUCT_TAG_VISIT_1",
        "IDEA_PIN_PRODUCT_TAG_VISIT_2",
        "IMPRESSION_1",
        "IMPRESSION_1_GROSS",
        "IMPRESSION_2",
        "INAPP_CHECKOUT_COST_PER_ACTION",
        "OUTBOUND_CLICK_1",
        "OUTBOUND_CLICK_2",
        "PAGE_VISIT_COST_PER_ACTION",
        "PAGE_VISIT_ROAS",
        "PAID_IMPRESSION",
        "PIN_ID",  #
        "PIN_PROMOTION_ID",  #
        "REPIN_1",  #
        "REPIN_2",  #
        "REPIN_RATE",
        "SPEND_IN_DOLLAR",
        "SPEND_IN_MICRO_DOLLAR",
        "TOTAL_CHECKOUT",
        "TOTAL_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_CLICKTHROUGH",
        "TOTAL_CLICK_ADD_TO_CART",  #
        "TOTAL_CLICK_CHECKOUT",
        "TOTAL_CLICK_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_CLICK_LEAD",  #
        "TOTAL_CLICK_SIGNUP",
        "TOTAL_CLICK_SIGNUP_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_CONVERSIONS",
        "TOTAL_CUSTOM",  #
        "TOTAL_ENGAGEMENT",  #
        "TOTAL_ENGAGEMENT_CHECKOUT",
        "TOTAL_ENGAGEMENT_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_ENGAGEMENT_LEAD",  #
        "TOTAL_ENGAGEMENT_SIGNUP",
        "TOTAL_ENGAGEMENT_SIGNUP_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_IDEA_PIN_PRODUCT_TAG_VISIT",
        "TOTAL_IMPRESSION_FREQUENCY",
        "TOTAL_IMPRESSION_USER",
        "TOTAL_LEAD",  #
        "TOTAL_OFFLINE_CHECKOUT",  #
        "TOTAL_PAGE_VISIT",
        "TOTAL_REPIN_RATE",
        "TOTAL_SIGNUP",
        "TOTAL_SIGNUP_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_VIDEO_3SEC_VIEWS",
        "TOTAL_VIDEO_AVG_WATCHTIME_IN_SECOND",
        "TOTAL_VIDEO_MRC_VIEWS",
        "TOTAL_VIDEO_P0_COMBINED",
        "TOTAL_VIDEO_P100_COMPLETE",
        "TOTAL_VIDEO_P25_COMBINED",
        "TOTAL_VIDEO_P50_COMBINED",
        "TOTAL_VIDEO_P75_COMBINED",
        "TOTAL_VIDEO_P95_COMBINED",
        "TOTAL_VIEW_ADD_TO_CART",  #
        "TOTAL_VIEW_CHECKOUT",
        "TOTAL_VIEW_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_VIEW_LEAD",  #
        "TOTAL_VIEW_SIGNUP",
        "TOTAL_VIEW_SIGNUP_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_WEB_CHECKOUT",
        "TOTAL_WEB_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_WEB_CLICK_CHECKOUT",
        "TOTAL_WEB_CLICK_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_WEB_ENGAGEMENT_CHECKOUT",
        "TOTAL_WEB_ENGAGEMENT_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "TOTAL_WEB_SESSIONS",  #
        "TOTAL_WEB_VIEW_CHECKOUT",
        "TOTAL_WEB_VIEW_CHECKOUT_VALUE_IN_MICRO_DOLLAR",
        "VIDEO_3SEC_VIEWS_2",
        "VIDEO_LENGTH",  #
        "VIDEO_MRC_VIEWS_2",
        "VIDEO_P0_COMBINED_2",
        "VIDEO_P100_COMPLETE_2",
        "VIDEO_P25_COMBINED_2",
        "VIDEO_P50_COMBINED_2",
        "VIDEO_P75_COMBINED_2",
        "VIDEO_P95_COMBINED_2",
        "WEB_CHECKOUT_COST_PER_ACTION",
        "WEB_CHECKOUT_ROAS",
        "WEB_SESSIONS_1",  #
        "WEB_SESSIONS_2",  #
    ]
    return ",".join(analytics_columns)


def to_datetime_str(date: datetime) -> str:
    """
    Returns the formated datetime string.
    :: Output example: '2021-07-15' FORMAT : "%Y-%m-%d"
    """
    return date.strftime("%Y-%m-%d")



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from typing import Mapping, Type, Union

from airbyte_cdk.models import FailureType
from airbyte_cdk.sources.streams.http.error_handlers.default_error_mapping import DEFAULT_ERROR_MAPPING
from airbyte_cdk.sources.streams.http.error_handlers.response_models import ErrorResolution, ResponseAction


PINTEREST_AD_ACCOUNT_DOCS_URL = "https://developers.pinterest.com/docs/api/v5/ad_accounts-get"
ERROR_MESSAGE = "Invalid Ad Account ID: {account_id}. Access denied. Check your permissions or visit {docs_url} for guidance."


def get_pinterest_ad_account_error_mapping(account_id: str) -> Mapping[Union[int, str, Type[Exception]], ErrorResolution]:
    """
    Maps API response errors to messages for Pinterest Ad Account validation.
    """
    return DEFAULT_ERROR_MAPPING | {
        403: ErrorResolution(
            response_action=ResponseAction.FAIL,
            failure_type=FailureType.config_error,
            error_message=ERROR_MESSAGE.format(account_id=account_id, docs_url=PINTEREST_AD_ACCOUNT_DOCS_URL),
        ),
        400: ErrorResolution(
            response_action=ResponseAction.FAIL,
            failure_type=FailureType.config_error,
            error_message=ERROR_MESSAGE.format(account_id=account_id, docs_url=PINTEREST_AD_ACCOUNT_DOCS_URL),
        ),
    }



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from abc import ABC
from datetime import datetime
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Union

import pendulum
import requests

from airbyte_cdk import AirbyteTracedException, BackoffStrategy
from airbyte_cdk.models import FailureType, SyncMode
from airbyte_cdk.sources.declarative.requesters.error_handlers.backoff_strategies import WaitTimeFromHeaderBackoffStrategy
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.http import HttpStream, HttpSubStream
from airbyte_cdk.sources.streams.http.error_handlers import ErrorHandler, ErrorResolution, HttpStatusErrorHandler, ResponseAction
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer
from source_pinterest.components.error_mapping import get_pinterest_ad_account_error_mapping

from .utils import get_analytics_columns, to_datetime_str


# For Pinterest analytics streams rate limit is 300 calls per day / per user.
# once hit - response would contain `code` property with int.
MAX_RATE_LIMIT_CODE = 8


class NonJSONResponse(Exception):
    pass


class RateLimitExceeded(Exception):
    pass


class PinterestErrorHandler(ErrorHandler):
    def __init__(self, logger: logging.Logger, stream_name: str) -> None:
        self._logger = logger
        self._stream_name = stream_name

    @property
    def max_retries(self) -> Optional[int]:
        """
        Default value from HttpStream before the migration
        """
        return 5

    @property
    def max_time(self) -> Optional[int]:
        """
        Default value from HttpStream before the migration
        """
        return 60 * 10

    def _handle_unknown_error(self, response: Optional[Union[requests.Response, Exception]]) -> Optional[ErrorResolution]:
        """
        Error handling could potentially be improved. For example, connection errors are probably transient as we could retry.
        """
        if isinstance(response, Exception):
            return ErrorResolution(
                ResponseAction.FAIL,
                FailureType.system_error,
                f"Failed because of the following error: {response}",
            )
        return None

    def interpret_response(self, response: Optional[Union[requests.Response, Exception]]) -> ErrorResolution:
        unhandled_error_resolution = self._handle_unknown_error(response)
        if unhandled_error_resolution:
            return unhandled_error_resolution

        try:
            resp = response.json()
        except requests.exceptions.JSONDecodeError:
            raise NonJSONResponse(f"Received unexpected response in non json format: '{response.text}'")

        # when max rate limit exceeded, we should skip the stream.
        if response.status_code == requests.codes.too_many_requests and (
            isinstance(resp, dict) and resp.get("code", 0) == MAX_RATE_LIMIT_CODE
        ):
            self._logger.error(f"For stream {self._stream_name} Max Rate Limit exceeded.")
            return ErrorResolution(
                ResponseAction.FAIL,
                FailureType.transient_error,
                "Max Rate Limit exceeded",
            )
        elif response.status_code == requests.codes.too_many_requests or 500 <= response.status_code < 600:
            return ErrorResolution(
                ResponseAction.RETRY,
                FailureType.transient_error,
                f"Failed after retrying on status code {response.status_code}: {response.content}",
            )
        elif not response.ok:
            return ErrorResolution(
                response_action=ResponseAction.FAIL,
                failure_type=FailureType.system_error,
                error_message=f"Response status code: {response.status_code}. Unexpected error. Failed.",
            )

        return ErrorResolution(ResponseAction.SUCCESS)


def _create_retry_after_backoff_strategy() -> BackoffStrategy:
    return WaitTimeFromHeaderBackoffStrategy(header="Retry-After", max_waiting_time_in_seconds=600, parameters={}, config={})


class PinterestStream(HttpStream, ABC):
    url_base = "https://api.pinterest.com/v5/"
    primary_key = "id"
    data_fields = ["items"]
    raise_on_http_errors = True
    max_rate_limit_exceeded = False
    transformer = TypeTransformer(TransformConfig.DefaultSchemaNormalization)

    def __init__(self, config: Mapping[str, Any]) -> None:
        super().__init__(authenticator=config["authenticator"])
        self.config = config

    def get_backoff_strategy(self) -> Optional[Union[BackoffStrategy, List[BackoffStrategy]]]:
        return _create_retry_after_backoff_strategy()

    def get_error_handler(self) -> ErrorHandler:
        return PinterestErrorHandler(self.logger, self.name)

    @property
    def start_date(self) -> str:
        return self.config["start_date"]

    @property
    def window_in_days(self) -> int:
        return 30  # Set window_in_days to 30 days date range

    def next_page_token(self, response: requests.Response) -> Optional[Mapping[str, Any]]:
        next_page = response.json().get("bookmark", {}) if self.data_fields else {}

        if next_page:
            return {"bookmark": next_page}

    def request_params(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, any] = None, next_page_token: Mapping[str, Any] = None
    ) -> MutableMapping[str, Any]:
        return next_page_token or {}

    def parse_response(self, response: requests.Response, stream_state: Mapping[str, Any], **kwargs) -> Iterable[Mapping]:
        """
        Parsing response data with respect to Rate Limits.
        """
        data = response.json()

        if not self.max_rate_limit_exceeded:
            for data_field in self.data_fields:
                data = data.get(data_field, [])

            for record in data:
                yield record


class AdAccounts(PinterestStream):
    use_cache = True
    is_resumable = False

    def path(self, **kwargs) -> str:
        return "ad_accounts"


class PinterestSubStream(HttpSubStream):
    def stream_slices(
        self, sync_mode: SyncMode, cursor_field: List[str] = None, stream_state: Mapping[str, Any] = None
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        parent_stream_slices = self.parent.stream_slices(
            sync_mode=SyncMode.full_refresh, cursor_field=cursor_field, stream_state=stream_state
        )
        # iterate over all parent stream_slices
        for stream_slice in parent_stream_slices:
            parent_records = self.parent.read_records(sync_mode=SyncMode.full_refresh, stream_slice=stream_slice)

            # iterate over all parent records with current stream_slice
            for record in parent_records:
                yield {"parent": record, "sub_parent": stream_slice}


class IncrementalPinterestStream(PinterestStream, ABC):
    def get_updated_state(self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]) -> Mapping[str, Any]:
        default_value = self.start_date.format("YYYY-MM-DD")
        latest_state = latest_record.get(self.cursor_field, default_value)
        current_state = current_stream_state.get(self.cursor_field, default_value)
        latest_state_is_numeric = isinstance(latest_state, int) or isinstance(latest_state, float)

        if latest_state_is_numeric and isinstance(current_state, str):
            current_state = datetime.strptime(current_state, "%Y-%m-%d").timestamp()

        return {self.cursor_field: max(latest_state, current_state)}

    def stream_slices(
        self, sync_mode: SyncMode, cursor_field: List[str] = None, stream_state: Mapping[str, Any] = None
    ) -> Iterable[Optional[Mapping[str, any]]]:
        """
        Override default stream_slices CDK method to provide date_slices as page chunks for data fetch.
        Returns list of dict, example: [{
            "start_date": "2020-01-01",
            "end_date": "2021-01-02"
            },
            {
            "start_date": "2020-01-03",
            "end_date": "2021-01-04"
            },
            ...]
        """

        start_date = self.start_date
        end_date = pendulum.now()

        # determine stream_state, if no stream_state we use start_date
        if stream_state:
            state = stream_state.get(self.cursor_field)

            state_is_timestamp = isinstance(state, int) or isinstance(state, float)
            if state_is_timestamp:
                state = str(datetime.fromtimestamp(state).date())

            start_date = pendulum.parse(state)

        # use the lowest date between start_date and self.end_date, otherwise API fails if start_date is in future
        start_date = min(start_date, end_date)
        date_slices = []

        while start_date < end_date:
            # the amount of days for each data-chunk beginning from start_date
            end_date_slice = (
                end_date if end_date.subtract(days=self.window_in_days) < start_date else start_date.add(days=self.window_in_days)
            )
            date_slices.append({"start_date": to_datetime_str(start_date), "end_date": to_datetime_str(end_date_slice)})

            # add 1 day for start next slice from next day and not duplicate data from previous slice end date.
            start_date = end_date_slice.add(days=1)

        return date_slices


class IncrementalPinterestSubStream(IncrementalPinterestStream):
    cursor_field = "updated_time"

    def __init__(self, parent: Stream, with_data_slices: bool = True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.parent = parent
        self.with_data_slices = with_data_slices

    def stream_slices(
        self, sync_mode: SyncMode, cursor_field: List[str] = None, stream_state: Mapping[str, Any] = None
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        date_slices = super().stream_slices(sync_mode, cursor_field, stream_state) if self.with_data_slices else [{}]
        parents_slices = PinterestSubStream.stream_slices(self, sync_mode, cursor_field, stream_state) if self.parent else [{}]

        for parents_slice in parents_slices:
            for date_slice in date_slices:
                parents_slice.update(date_slice)

                yield parents_slice


def _lookback_date_limit_reached(response: requests.Response) -> bool:
    """
    After few consecutive requests, analytics API return bad request error with 'You can only get data from the last 90 days' error
    message. But with next request all working good. So, we wait 1 sec and request again if we get this issue.
    """
    if isinstance(response.json(), dict):
        return response.json().get("code", 0) and response.status_code == 400
    return False


class PinterestAnalyticsErrorHandler(ErrorHandler):
    def __init__(self, logger: logging.Logger, stream_name: str) -> None:
        self._decorated = PinterestErrorHandler(logger, stream_name)

    @property
    def max_retries(self) -> Optional[int]:
        """
        Default value from HttpStream before the migration
        """
        return 5

    @property
    def max_time(self) -> Optional[int]:
        """
        Default value from HttpStream before the migration
        """
        return 60 * 10

    def interpret_response(self, response: Optional[Union[requests.Response, Exception]]) -> ErrorResolution:
        if isinstance(response, requests.Response) and _lookback_date_limit_reached(response):
            return ErrorResolution(
                ResponseAction.RETRY,
                FailureType.transient_error,
                f"Analytics API returns bad request error when under load. This error should be retried after a second. If this error message appears, it means the Analytics API did not recover or there might be a bigger issue so please contact the support team.",
            )

        return self._decorated.interpret_response(response)


class AnalyticsApiBackoffStrategyDecorator(BackoffStrategy):
    def __init__(self) -> None:
        self._decorated = _create_retry_after_backoff_strategy()

    def backoff_time(
        self, response_or_exception: Optional[Union[requests.Response, requests.RequestException]], **kwargs
    ) -> Optional[float]:
        if isinstance(response_or_exception, requests.Response) and _lookback_date_limit_reached(response_or_exception):
            return 1
        return self._decorated.backoff_time(response_or_exception)


class PinterestAnalyticsStream(IncrementalPinterestSubStream):
    primary_key = None
    cursor_field = "DATE"
    data_fields = []
    granularity = "DAY"
    analytics_target_ids = None

    def get_backoff_strategy(self) -> Optional[Union[BackoffStrategy, List[BackoffStrategy]]]:
        return AnalyticsApiBackoffStrategyDecorator()

    def get_error_handler(self) -> ErrorHandler:
        return PinterestAnalyticsErrorHandler(self.logger, self.name)

    def request_params(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, any] = None, next_page_token: Mapping[str, Any] = None
    ) -> MutableMapping[str, Any]:
        params = super().request_params(stream_state, stream_slice, next_page_token)
        params.update(
            {
                "start_date": stream_slice["start_date"],
                "end_date": stream_slice["end_date"],
                "granularity": self.granularity,
                "columns": get_analytics_columns(),
            }
        )

        if self.analytics_target_ids:
            params.update({self.analytics_target_ids: stream_slice["parent"]["id"]})

        return params


class AdAccountValidationStream(PinterestStream):
    """
    This stream checks if the given ad_account_id exists by making an API request.

    - If the account exists, the API returns 200.
    - If the account does not exist, it returns an error.
    """

    def __init__(self, config: Mapping[str, Any]) -> None:
        self.account_id = config.get("account_id")
        super().__init__(config)

    def path(self, **kwargs) -> str:
        """Returns the API endpoint for validating an account ID."""
        return f"ad_accounts/{self.account_id}"

    def get_error_mapping(self):
        """Maps API errors using predefined error messages."""
        return get_pinterest_ad_account_error_mapping(self.account_id)

    def get_error_handler(self) -> HttpStatusErrorHandler:
        return HttpStatusErrorHandler(logger=self.logger, error_mapping=self.get_error_mapping())

    def parse_response(self, response: requests.Response, **kwargs) -> Iterable[Mapping[str, Any]]:
        """Returns the response if the account exists, otherwise throws an error."""
        if response.status_code == 200:
            yield response.json()
        else:
            yield from []



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from enum import Enum
from typing import List, Optional

from pydantic.v1 import BaseModel


class ReportStatus(str, Enum):
    """Enum Class to define the possible status of a report"""

    DOES_NOT_EXIST = "DOES_NOT_EXIST"
    EXPIRED = "EXPIRED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"
    FINISHED = "FINISHED"
    IN_PROGRESS = "IN_PROGRESS"


class ReportStatusDetails(BaseModel):
    """Model to capture details of the report status"""

    report_status: ReportStatus
    url: Optional[str]
    size: Optional[int]


class ReportInfo(BaseModel):
    """Model to capture details of the report info"""

    report_status: ReportStatus
    token: str
    message: Optional[str]
    metrics: Optional[List[dict]]
